/*
==mysql==
  Connection object for mysql, roughly following the python database API specification v2.0

====Constructors====
;mysql() 
:Creates new, unconnected object

;mysql(...)
:Arguments are passed to connect.


====Example====
 var db=new mysql("localhost","user","password","database");
 // or
 var db=open("mysql://user:password@localhost/database");
 // same result

 var people=db.query("SELECT * FROM people");
 var i;
 for (i in people) {
 	print(people[i].id, people[i].name,"\n");
 }

 var name=db.query("SELECT name FROM people");
 print(name[0],"\n");

 db.query("INSERT INTO people (id, name) VALUES (:id, :name)", {id: 42, name:"John"});
*/

mysql=function() {
	this.cursors=[];
	this.clones=[];

	if (arguments.length>0) this.connect.apply(this,arguments);
}

mysql.open_uri=function(uri) {
	if (uri.path!=undefined)
		uri.path=uri.path.substr(1);
	if (uri.qry==undefined) uri.qry={};
	return new mysql(uri.host, uri.username, uri.password, uri.path,
		uri.port, uri.qry.unix_socket, uri.qry.clientflag);
}

mysql.prototype=new dbapi20;

/*
====query(qry, [parameters])====

Creates a cursor, passes the query to the server and
fetches all rows in the result using cursor.fetchall().
Closes the cursor.

Inherited from dbapi20
*/


/*
====connect([host], [user], [passwd], [db], [port], [unix_socket], [clientflag])====
*/

mysql.prototype.connect=function(host, user, passwd, db, port, unix_socket, clientflag) {
	if (host==undefined) host=null;
	if (user==undefined) user=null;
	if (passwd==undefined) passwd=null;
	if (db==undefined) db=null;
	if (port==undefined) port=null;
	if (unix_socket==undefined) unix_socket=null;
	if (clientflag==undefined) clientflag=0;

	this.connparam=[host,user,passwd,db,port,unix_socket,clientflag];
	this.using_result=false;

	this.mysql=_.mysql_init(null);
	if (null==_.mysql_real_connect(this.mysql, host, user, passwd, db, port, unix_socket, clientflag)) {
		this.throwerror();
	}

	this.autocommit(false);
}

/**/

// Used for multiple simultaneous use_result queries (private)

mysql.prototype.new_clone=function() {
	var ret=new mysql;
	ret.connect.apply(ret,this.connparam);
	this.clones.push(ret);
	return ret;
}

/**/

// Used for reporting errors from the mysql api (private)

mysql.prototype.throwerror=function() {
	throw("mysql: "+_.mysql_error(this.mysql).string());
}

/**/

// Used for multiple simultaneous use_result queries (private)

mysql.prototype.get_clone=function() {
	var i;
	for (i in this.clones) {
		if (!this.clones[i].using_result) {
			return this.clones[i];
		}
	}
	return this.new_clone();
}

/*
====close()====

            Close the connection The connection will be unusable from this point
            forward; an Error (or subclass) exception will be raised
            if any operation is attempted with the connection. The
            same applies to all cursor objects trying to use the
            connection.  Note that closing a connection without
            committing the changes first will cause an implicit
            rollback to be performed.
*/

mysql.prototype.close=function() {
	this.rollback();
	var i;
	for (i in this.cursors) {
		this.cursors[i].close();
	}
	this.cursors=[];

	for (i in this.clones) {
		this.clones[i].close();
	}
	this.clones=[];

	_.mysql_close(this.mysql);
	delete this.mysql;
}

/*
====autocommit(Boolean mode)====

Turns autocommit on or off. Is initially off.
*/

mysql.prototype.autocommit=function(mode) {
	if (_.mysql_autocommit(this.mysql,mode)) {
		this.throwerror();
	}
}

/*
====commit()====
            Commit any pending transaction to the database. Note that
            if the database supports an auto-commit feature, this must
            be initially off. An interface method may be provided to
            turn it back on.
            
            Database modules that do not support transactions should
            implement this method with void functionality.
*/

mysql.prototype.commit=function() {
	if (_.mysql_commit(this.mysql)) {
		this.throwerror();
	}
}

/*
====rollback()====
          
            This method is optional since not all databases provide
            transaction support.
            
            In case a database does provide transactions this method
            causes the the database to roll back to the start of any
            pending transaction.  Closing a connection without
            committing the changes first will cause an implicit
            rollback to be performed.
*/

mysql.prototype.rollback=function() {
	if (_.mysql_rollback(this.mysql)) {
		this.throwerror();
	}
}

/*
====cursor()====

            Return a new mysql.cursor object using the connection.  If the
            database does not provide a direct cursor concept, the
            module will have to emulate cursors using other means to
            the extent needed by this specification.
*/

mysql.prototype.cursor=function() {
	var ret=new mysql.cursor(this);
	this.cursors.push(ret);
	return ret;
}

/*
    mysql.prototype.messages

        Same as cursor.messages except that the messages in the list
        are connection oriented.

        The list is cleared automatically by all standard connection
        methods calls (prior to executing the call) to avoid excessive
        memory usage and can also be cleared by executing "delete
        connection.messages".
*/

/*
===mysql.cursor===

Represents a MySQL result set
*/


mysql.cursor=function(connection) {
	this.connection=connection;
}

mysql.cursor.prototype=new dbapi20.cursor;

/*
====description()====

            This function returns a sequence of 7-item
            objects.  Each of these objects contains information
            describing one result column: (name, type_code,
            display_size, internal_size, precision, scale,
            null_ok). The first two items (name and type_code) are
            mandatory, the other five are optional and must be set to
            None if meaningfull values are not provided.

            This attribute will be None for operations that
            do not return rows or if the cursor has not had an
            operation invoked via the executeXXX() method yet.
            
            The type_code can be interpreted by comparing it to the
            Type Objects specified in the section below.
*/

mysql.cursor.prototype.description=function() {
	if (this.result==null) {
		if (!this.initfetch_use()) {
			return null;
		}
	}

	var ret=[];
	var field;
	_.mysql_field_seek(this.result,0);
	while ((field=_.mysql_fetch_field(this.result))!=null) {
		ret.push({
			name: field.name.string(field.name_length),
			type_code: field.member("type").get(),
			display_size: 0,
			internal_size: 0,
			precision: field.decimals,
			scale: 0,
			null_ok: 0
			});
	}
	return ret;
}

/*
====rowcount()====

            This function returns the number of rows that
            the last executeXXX() produced (for DQL statements like
            'select') or affected (for DML statements like 'update' or
            'insert').
            
            The function returns null in case no executeXXX() has been
            performed on the cursor or the rowcount of the last
            operation is not determinable by the interface. [7]

*/

mysql.cursor.prototype.rowcount=function() {
	return _.mysql_num_rows(this.result);
}

/*
====callproc(procname[,parameters])====

            (This method is optional since not all databases provide
            stored procedures. [3])
            
            Call a stored database procedure with the given name. The
            sequence of parameters must contain one entry for each
            argument that the procedure expects. The result of the
            call is returned as modified copy of the input
            sequence. Input parameters are left untouched, output and
            input/output parameters replaced with possibly new values.
            
            The procedure may also provide a result set as
            output. This must then be made available through the
*/

mysql.cursor.prototype.callproc=function(procname,parameters) {
}

/*
====close()====

          
            Close the cursor. The cursor will be unusable from this point
            forward; an Error (or subclass) exception will be raised
            if any operation is attempted with the cursor.
*/

mysql.cursor.prototype.close=function() {
	this.free();
}


mysql.cursor.prototype.param_subst=function(operation, paramname, val) {
	var parampos=operation.indexOf(':'+paramname);
	if (parampos==-1) throw("mysql: Unknown parameter "+i);
	switch(typeof(val)) {
	case "string": // Escape and quote string
		var to=new Pointer(val.length*2+1);
		var len=_.mysql_real_escape_string(this.connection.mysql, to, val, val.length);
		val="'"+to.string(len)+"'";
		break;
	case "object":
		if (val instanceof Date) {
			val="'"+val.getFullYear()+"-"+val.getMonth()+"-"+val.getDate+" "+
				val.getHours()+":"+val.getMinutes+":"+val.getSeconds+"'";
		}
	}
	return operation.substr(0,parampos)+val+operation.substr(parampos+paramname.length+1);
}

/*
        ====execute(operation[,parameters])====
          
            Prepare and execute a database operation (query or
            command).  Parameters may be provided as sequence or
            mapping and will be bound to variables in the operation.
	String parameters will be escaped so as to be binary-safe.
	Unicode strings are not supported. Date values will be
	converted to MySQL's date format.

            Return values are not defined.
*/

mysql.cursor.prototype.execute=function(operation,parameters) {
	this.free();
	if (this.connection.using_result==true) {
		this.connection=this.connection.get_clone();
	}
	var argno;
	for (argno=1; argno<arguments.length; argno++) {
		if (typeof(arguments[argno])=="object") {
			var i;
			for (i in parameters) {
				i=String(i);
				operation=this.param_subst(operation,i,parameters[i]);
			}
		} else {
			operation=this.param_subst(operation,"",arguments[argno]);
		}
	}
	var res=_.mysql_real_query(this.connection.mysql, operation, operation.length);
	if (res) this.connection.throwerror();
}

/*
*/

mysql.cursor.prototype.result=null;

/*
        ====executemany(operation,array_of_parameters)====
          
            Implemented using multiple calls to execute.

            The same comments as for execute() also apply accordingly
            to this method.
            
            Return values are not defined.

Inherited from dbapi20.cursor
*/

/*
        ====fetchone() ====
          
            Fetch the next row of a query result set, returning a
            single value or object, or null when no more data is
            available.

	If the query results in zero or more than one columns, the return value is
	an object which contains properties with the same name as the column names.
	Values are converted into strings, numbers or dates.

	If the query results in a single column, that column is returned as a value
	instead of a single-property object.

Inherited from dbapi20.cursor
*/


/*
        ====fetchmany([size=cursor.arraysize])====
          
            Fetch the next set of rows of a query result, returning an
            array of objects. null
            is returned when no more rows are available.
            
            The number of rows to fetch per call is specified by the
            parameter.  If it is not given, the cursor's arraysize
            determines the number of rows to be fetched. The method
            tries to fetch as many rows as indicated by the size
            parameter. If this is not possible due to the specified
            number of rows not being available, fewer rows may be
            returned.

	If size=-1, will fetch all remaining rows
            
*/

mysql.cursor.prototype.fetchmany=function(size) {
	if (arguments.length<1) size=this.arraysize;
	var i, j, outval, val, row, outrow;
	var ret=[];
	if (this.result==null) {
		if (!this.initfetch_use()) {
			return null;
		}
	}

	var nfields=this.field_names.length;

	for (i=0; size==-1 || i<size; i++) {
		row=_.mysql_fetch_row(this.result);
		if (row==null) break;

		var lengths=_.mysql_fetch_lengths(this.result);

		if (nfields!=1) outrow=new Object;
		for (j=0; j<this.field_names.length; j++) {
			val=row[j].string(lengths[j]);
			switch(this.field_types[j]) {
			case _.MYSQL_TYPE_DATE:
				var date=val.match(/([0-9]*)-([0-9]*)-([0-9]*)/);
				outval=new Date(date[1],date[2],date[3]);
				break;
			case _.MYSQL_TYPE_TIME:
				var date=val.match(/([0-9]*):([0-9]*):([0-9]*)/);
				outval=new Date(0,0,0,date[1],date[2],date[3]);
				break;
			case _.MYSQL_TYPE_DATETIME:
				var date=val.match(/([0-9]*)-([0-9]*)-([0-9]*) ([0-9]*):([0-9]*):([0-9]*)/);
				outval=new Date(date[1],date[2],date[3],date[4],date[5],date[6]);
				break;
			case _.MYSQL_TYPE_DECIMAL:
			case _.MYSQL_TYPE_TINY:
			case _.MYSQL_TYPE_SHORT:
			case _.MYSQL_TYPE_LONG:
			case _.MYSQL_TYPE_FLOAT:
			case _.MYSQL_TYPE_DOUBLE:
			case _.MYSQL_TYPE_LONGLONG:
			case _.MYSQL_TYPE_INT24:
				outval=Number(val);
				break;
			default:
				outval=val;
			}
			if (nfields!=1) outrow[this.field_names[j]]=outval;
		}
		if (nfields==1) outrow=outval;

		this.rownumber++;
		ret.push(outrow);
	}

	if (ret.length==0) return null;
	return ret;
}

/*           
        ====fetchall() ====

            Fetch all (remaining) rows of a query result, returning
            them as an array of objects.
*/

mysql.cursor.prototype.fetchall=function() {
	if (this.result==null) {
		if (!this.initfetch_store()) {
			return null;
		}
	}

	var ret=this.fetchmany(-1);

	this.free();
	return ret;
}

/**/

mysql.cursor.prototype.free=function() {
	if (this.result) {
		_.mysql_free_result(this.result);
		this.result=null;
		this.connection.using_result=false;
	}
}

/**/

mysql.cursor.prototype.initfetch_store=function() {
	this.free();
	this.result=_.mysql_store_result(this.connection.mysql);
	if (this.result==null) {
		if (_.mysql_field_count(this.connection.mysql)==0) return false;
		this.connection.throwerror();
	}
	this.get_fields();
	this.rownumber=0;
	return true;
}

/**/

mysql.cursor.prototype.initfetch_use=function() {
	this.free();
	this.result=_.mysql_use_result(this.connection.mysql);
	if (this.result==null) {
		if (_.mysql_field_count(this.connection.mysql)==0) return false;
		this.connection.throwerror();
	}
	this.connection.using_result=true;
	this.get_fields();
	this.rownumber=0;
	return true;
}

/**/

mysql.cursor.prototype.get_fields=function() {
	this.field_names=[];
	this.field_types=[];
	var field;
	while ((field=_.mysql_fetch_field(this.result))!=null) {
		this.field_types.push(field.member("type").get());
		this.field_names.push(field.name.string(field.name_length));
	}
}

/*
        ====nextset() ====
          
            (This method is optional since not all databases support
            multiple result sets. [3])
            
            This method will make the cursor skip to the next
            available set, discarding any remaining rows from the
            current set.
            
            If there are no more sets, the method returns
            None. Otherwise, it returns a true value and subsequent
            calls to the fetch methods will return rows from the next
            result set.
            
            An Error (or subclass) exception is raised if the previous
            call to executeXXX() did not produce any result set or no
            call was issued yet.
*/

mysql.cursor.prototype.nextset=function() {
	this.free();
	this.result=_.mysql_next_result(this.connection.mysql);
	if (this.result==null) this.connection.throwerror();
	this.get_fields();
	this.rownumber=0;
}

/*
        ====arraysize====
          
            This read/write attribute specifies the number of rows to
            fetch at a time with fetchmany(). It defaults to 1 meaning
            to fetch a single row at a time.
            
            Implementations must observe this value with respect to
            the fetchmany() method, but are free to interact with the
            database a single row at a time. It may also be used in
            the implementation of executemany().

Inherited from dbapi20.cursor
*/

/*            
        ====setinputsizes(sizes)====
          
            This can be used before a call to executeXXX() to
            predefine memory areas for the operation's parameters.
            
            sizes is specified as a sequence -- one item for each
            input parameter.  The item should be a Type Object that
            corresponds to the input that will be used, or it should
            be an integer specifying the maximum length of a string
            parameter.  If the item is None, then no predefined memory
            area will be reserved for that column (this is useful to
            avoid predefined areas for large inputs).
            
            This method would be used before the executeXXX() method
            is invoked.
            
            Implementations are free to have this method do nothing
            and users are free to not use it.
*/

mysql.cursor.prototype.setinputsizes=function(sizes) {
}

/*
        ====setoutputsize(size[,column])====
          
            Set a column buffer size for fetches of large columns
            (e.g. LONGs, BLOBs, etc.).  The column is specified as an
            index into the result sequence.  Not specifying the column
            will set the default size for all large columns in the
            cursor.
            
            This method would be used before the executeXXX() method
            is invoked.
            
            Implementations are free to have this method do nothing
            and users are free to not use it.
*/

mysql.cursor.prototype.setoutputsize=function(size,column) {
}

/*
    ====rownumber====

        This read-only attribute is the current 0-based
        index of the cursor in the result set or null if the index cannot
        be determined.

        The index can be seen as index of the cursor in a sequence (the
        result set). The next fetch operation will fetch the row
        indexed by .rownumber in that sequence.

Inherited from dbapi20.cursor
*/


/*
   ====connection====

        This read-only attribute return a reference to the Connection
        object on which the cursor was created.

        The attribute simplifies writing polymorph code in
        multi-connection environments.
*/

/*
    ====scroll(value[,mode='relative'])====

        Scroll the cursor in the result set to a new position according
        to mode.

        If mode is 'relative' (default), value is taken as offset to
        the current position in the result set, if set to 'absolute',
        value states an absolute target position.

        An IndexError should be raised in case a scroll operation would
        leave the result set. In this case, the cursor position is left
        undefined (ideal would be to not move the cursor at all).

        Note: This method should use native scrollable cursors, if
        available , or revert to an emulation for forward-only
        scrollable cursors. The method may raise NotSupportedErrors to
        signal that a specific operation is not supported by the
        database (e.g. backward scrolling).

*/

mysql.cursor.prototype.scroll=function(value,mode) {
	switch(mode) {
	case "absolute":
		_.mysql_data_seek(this.result, value);
		this.rownumber=value;
		break;
	default:
		_.mysql_data_seek(this.result, this.rownumber+value);
		this.rownumber+=value;
		break;
	}
}

/*
    ====messages====

        This is an array object to which the interface appends
        errors for all messages
        which the interfaces receives from the underlying database for
        this cursor.

        The list is cleared by all standard cursor methods calls (prior
        to executing the call) except for the .fetchXXX() calls
        automatically to avoid excessive memory usage and can also be
        cleared by executing "delete cursor.messages".

        All error and warning messages generated by the database are
        placed into this array, so checking the list allows the user to
        verify correct operation of the method calls.

        The aim of this attribute is to eliminate the need for a
        Warning exception which often causes problems (some warnings
        really only have informational character).

*/

