/*
==os==

This module provides a more portable way of using operating system dependent functionality than importing a operating system dependent built-in module like [[posix]] or [[nt]]. 

This module searches for an operating system dependent built-in module like mac or posix and exports the same functions and data as found there. The design of all Python's built-in operating system dependent modules is such that as long as the same functionality is available, it uses the same interface; for example, the function [[os.stat]](path) returns stat information about path in the same format (which happens to have originated with the POSIX interface). 

Extensions peculiar to a particular operating system are also available through the os module, but using them is of course a threat to portability! 

Note that after the first time [[os]] is imported, there is no performance penalty in using functions from [[os]] instead of directly from the operating system dependent built-in module, so there should be no reason not to use [[os]]!
*/

/*
====exit(exit_code)====

Exits jsext with the specified exit code
*/

os.exit=_.exit;

/*
====fdopen(fd[, mode[, bufsize]]) ====

Return an open file object connected to the file descriptor ''fd''.  The mode and bufsize arguments have the same meaning as the corresponding arguments to the built-in [[open]]() function. Availability: Macintosh, Unix, Windows. 
*/

os.fdopen=function(fd, mode, bufsize) {
	if (!mode) mode="r";

	var ret=new file(_.fdopen(fd, mode), bufsize);
	ret.name="<fd "+fd+">";

	return ret;
}

/*
====popen(command[, mode[, bufsize]]) ====

Open a pipe to or from ''command''. The return value is an open file object connected to the pipe, which can be read or written depending on whether mode is 'r' (default) or 'w'. The bufsize argument has the same meaning as the corresponding argument to the built-in [[open]]() function. The exit status of the command (encoded in the format specified for [[os.wait]]()) is available as the return value of the close() method of the file object, except that when the exit status is zero (termination without errors), None is returned. Availability: Unix, Windows. 

Changed in version 2.0: This function worked unreliably under Windows in earlier versions of Python. This was due to the use of the _popen() function from the libraries provided with Windows. Newer versions of Python do not use the broken implementation from the Windows libraries. 

*/

os.popen=function(command, mode, bufsize) {
	if (mode==undefined) mode="r";
	else if (mode!="r" && mode!="w" && mode!="a") throw("os.popen: Invalid mode");

	var ret=new file(_.popen(command, mode), bufsize);
	ret.name="<command "+command+">";

	ret.close=function() {
		if (!this.closed) {
			var ret=_.pclose(this.fp);
			this.closed=true;
			if (ret>=256) ret/=256;
			return ret;
		}
	}

	return ret;
}

/*
====tmpfile() ====

Return a new file object opened in update mode ("w+"). The file has no directory entries associated with it and will be automatically deleted once there are no file descriptors for the file. Availability: Unix, Windows. 
*/


os.tmpfile=function() {
	return new file(_.tmpfile());
}


/*
abort() -> does not return!

Abort the interpreter immediately.  This 'dumps core' or otherwise fails
in the hardest way possible on the hosting operating system.
*/


os.abort=_.abort;

/*
====popen2(cmd[, mode[, bufsize]]) ====

Executes cmd as a sub-process. Returns the file objects (child_stdin, child_stdout). Availability: Unix, Windows. New in version 2.0. 
*/

os.popen2=function(command, mode, bufsize, variant) {
	if (!mode) mode="t";
	if (!variant) variant=2;

	var stdout=[0,0];
	var stdin=[0,0];
	if (variant==3) {
		var stderr=[0,0];
		_.pipe(stderr);
	}
	_.pipe(stdin);
	_.pipe(stdout);

	var pid=_.fork();
	if (pid==0) {
		// Child process
		_.close(stdin[1]);
		_.dup2(stdin[0],_.STDIN_FILENO);
		_.close(stdin[0]);
		_.close(stdout[0]);
		_.dup2(stdout[1],_.STDOUT_FILENO);
		_.close(stdout[1]);
		if (variant==3) {
			_.close(stderr[0]);
			_.dup2(stderr[1],_.STDERR_FILENO);
			_.close(stderr[1]);
		}
		if (variant==4) {
			_.dup2(_.STDOUT_FILENO,_.STDERR_FILENO);
		}
		var args=command.split(/[ \t]+/);
		args.push(null);
		_.execvp(args[0],args);
		os.exit()
	}

	var proc=new Object;
	proc.pid=pid;

	if (variant==3) {
		proc.filecount=3;
	} else {
		proc.filecount=2;
	}

	var ret=[os.fdopen(stdin[1], "w"+mode, bufsize),
		os.fdopen(stdout[0], "r"+mode, bufsize)];
	_.close(stdin[0]);
	_.close(stdout[1]);

	if (variant==3) {
		ret.push(os.fdopen(stderr[0], "r"+mode, bufsize));
		_.close(stderr[1]);
	}

	for (i in ret) {
		ret[i].proc=proc;
		ret[i].close=function() {
			if (!this.closed) {
				_.fclose(this.fp);
				this.closed=true;
				this.proc.filecount--;
				if (this.proc.filecount==0) {
					return os.waitpid(this.proc.pid);
				}
			}
		}
	}
	return ret;
}


/*
====popen3(cmd[, mode[, bufsize]]) ====

Executes cmd as a sub-process. Returns the file objects (child_stdin, child_stdout, child_stderr). Availability: Unix, Windows. New in version 2.0.
*/

os.popen3=function(command, mode, bufsize) {
	return popen2(command, mode, bufsize, 3);
}

/*
====popen4(cmd[, mode[, bufsize]]) ====

Executes cmd as a sub-process. Returns the file objects (child_stdin, child_stdout_and_stderr). Availability: Unix, Windows. New in version 2.0. 
*/

os.popen4=function(command, mode, bufsize, variant) {
	return popen2(command, mode, bufsize, 4);
}

os.O_RDONLY=_.O_RDONLY;
os.O_WRONLY=_.O_WRONLY;
os.O_RDWR=_.O_RDWR;
os.O_APPEND=_.O_APPEND;
os.O_ASYNC=_.O_ASYNC;
os.O_CREAT=_.O_CREAT;
os.O_DIRECT=_.O_DIRECT;
os.O_EXCL=_.O_EXCL;
os.O_LARGEFILE=_.O_LARGEFILE;
os.O_NOATIME=_.O_NOATIME;
os.O_NOCTTY=_.O_NOCTTY;
os.O_NOFOLLOW=_.O_NOFOLLOW;
os.O_NONBLOCK=_.O_NONBLOCK;
os.O_NDELAY=_.O_NDELAY;
os.O_SYNC=_.O_SYNC;
os.O_TRUNC=_.O_TRUNC;

/*
open(filename, flag [, mode=0777]) -> fd

Open a file (for low level IO).	
*/


os.open=function(file, flags, mode) {
	if (mode==undefined) mode=0777;
	var ret=_.open(file,flags,Math.floor(mode));
	if (ret==-1) throw("open: Can't open "+file);
	return ret;
}

/*
write(fd, string) -> byteswritten

Write a string to a file descriptor.
*/

os.write=function(fd, str) {
	return _.write(fd, str, str.length);
}

/*
close(fd) -> None

Close a file descriptor (for low level IO).
*/

os.close=_.close;

/*
dup(fd) -> fd2

Return a duplicate of a file descriptor.
*/

os.dup=_.dup;

/*
dup2(fd, fd2) -> None

Duplicate file descriptor.
*/

os.dup2=_.dup2;

/*
fdatasync(fildes) -> None

force write of file with filedescriptor to disk.
 does not force update of metadata.
*/

os.fdatasync=_.fdatasync;

/*
fpathconf(fd, name) -> integer

Return the configuration limit name for the file descriptor fd.
If there is no limit, return -1.
*/

os.fpathconf=function(fd, name) {
	return _.fpathconf(fd, C["_PC_"+name]);
}


/*
fsync(fildes) -> None

force write of file with filedescriptor to disk.
*/

os.fsync=_.fsync;

/*
ftruncate(fd, length) -> None
Truncate a file to a specified length.
*/

os.ftruncate=_.ftruncate;

/*
isatty(fd) -> Boolean
*/

os.isatty=_.isatty;

/*
lseek(fd, pos, how) -> newpos

Set the current position of a file descriptor.
*/

os.lseek=_.lseek;

/*
pipe() -> (read_end, write_end)

Create a pipe.
*/

os.pipe=_.pipe;

/*
read(fd, buffersize) -> string

Read a file descriptor.
*/

os.read=_.read;

/*
tcgetpgrp(fd) -> pgid

Return the process group associated with the terminal given by a fd.
*/

os.tcgetgrp=_.tcgetgrp;

/*
tcsetpgrp(fd, pgid) -> None

Set the process group associated with the terminal given by a fd.
*/

os.tcsetgrp=_.tcsetgrp;

/*
fork() -> pid

Fork a child process.
 
Return 0 to child process and PID of child to parent process.
*/

os.fork=_.fork;

/*
getpgrp() -> pgrp

Return the current process group id.
*/

os.getpgrp=_.getpgrp;

/*
getpid() -> pid

Return the current process id
*/

os.getpid=_.getpid;

/*
getppid() -> ppid

Return the parent's process id.
*/

os.getppid=_.getppid;

/*
getuid() -> uid

Return the current process's user id.
*/

os.getuid=_.getuid;

/*
umask(new_mask) -> old_mask

Set the current numeric umask and return the previous umask.
*/

os.umask=_.umask;

/*
ttyname(fd) -> String

Return the name of the terminal device connected to 'fd'.
*/

os.ttyname=function(fd) {
	var buf=new Pointer(512);
	if (_.ttyname_r(fd, buf, 512)) os.error("ttyname");
	return buf.string();
}

/*
	strerror(errnum) -> String
	Returns error text for a given error number
*/

os.strerror=function(errnum) {
	return _.strerror(errnum).string();
}


/**/

os.error=function(str) {
	throw(str+": "+os.strerror(os.errno()));
}

/*
kill(pid, sig="TERM") -> None
Kill a process with a signal. sig may be a string or a number.
*/

os.kill=function(pid, sig) {
	if (sig==undefined) sig="TERM";
	if (typeof(sig)=="string") sig=C["SIG"+sig];
	if (_.kill(pid,sig)) os.error("os.kill");
}

/*
getgroups() -> list of group IDs

Return list of supplemental group IDs for the process.

*/
os.getgroups=function() {
	var numgroups=_.getgroups(0,null);
	var ret=[];
	var i;
	for (i=0; i<numgroups; i++)
		ret.push(0);
	_.getgroups(numgroups,ret);
	return ret;
}

/*
getlogin() -> string

Return the actual login name.

*/

os.getlogin=function() {
	return _.getlogin().string();
}


os.symlink=_.symlink;
os.sysconf=_.sysconf;
os.tcgetpgrp=_.tcgetpgrp;
os.tcsetpgrp=_.tcsetpgrp;

/*
times()

Return an object with integer numbers indicating process times.
*/

os.times=function() {
	var t=_.New(_.struct.tms);
	var res=_.times(t);
	if (res==-1) {
		os.error("times");
	}
	
	var i,j;
	var ret=t.get();
	return ret;
}

/*
===access(pathname, [mode=F_OK])===

access  checks  whether the process would be allowed to read, write or test for existence of the file (or other file system object) whose name is
       pathname.  If pathname is a symbolic link permissions of the file referred to by this symbolic link are tested.

       mode is a mask consisting of one or more of os.R_OK, os.W_OK, os.X_OK and os.F_OK.

       R_OK, W_OK and X_OK request checking whether the file exists and has read, write and  execute  permissions,  respectively.   F_OK  just  requests
       checking for the existence of the file.

       The  tests depend on the permissions of the directories occurring in the path to the file, as given in pathname, and on the permissions of direc-
       tories and files referred to by symbolic links encountered on the way.

       The check is done with the process's real UID and GID, rather than with the effective IDs as is done when actually attempting an operation.  This
       is to allow set-UID programs to easily determine the invoking user's authority.

       Only  access  bits are checked, not the file type or contents.  Therefore, if a directory is found to be "writable," it probably means that files
       can be created in the directory, and not that the directory can be written as a file.  Similarly, a DOS file may be found to be "executable," but
       the execve(2) call will still fail.

       If  the  process has appropriate privileges, an implementation may indicate success for X_OK even if none of the execute file permission bits are
       set.

====RETURN VALUE====

       On success (all requested permissions granted), true is returned.  On error (at least one bit in mode asked for a permission that is  denied,  or
       some other error occurred), false is returned, and errno is set appropriately.

*/

os.access=function(pathname, mode) {
	if (mode==undefined) mode=os.F_OK;
	var ret=_.access(pathname,mode);
	return ret==0;
}

os.R_OK=_.R_OK;
os.W_OK=_.W_OK;
os.X_OK=_.X_OK;
os.F_OK=_.F_OK;

/*
chdir(path) -> None

Change the current working directory to the specified path.
*/

os.chdir=_.chdir;

/*
fchdir(fd) -> None

Change the current working directory to the specified file descriptor, which should be an open directory
*/

os.fchdir=_.fchdir;

/*
chroot(path) -> None

Change the root of the process' file system to the given path
*/

os.chroot=_.chroot;

/*
chmod(path, mode) -> None

Change the access permissions of a file.
*/

os.chmod=_.chmod;

/*
chown(path, uid, gid) -> None

Change the owner and group id of path to the numeric uid and gid.
*/

os.chown=_.chown;

/*
lchown(path, uid, gid) -> None

Change the owner and group id of path to the numeric uid and gid.
Similar to chown, only that this version does *not* follow
symbolic links.
*/

os.lchown=_.lchown;

/*
link(src, dst) -> None

Create a hard link to a file.
*/

os.link=_.link;

/*
pathconf(path, name) -> integer

Return the configuration limit name for the file or directory path.
If there is no limit, return -1.
*/

os.pathconf=function(path, name) {
	return _.pathconf(path, C["_PC_"+name]);
}


/*
getegid() -> gid

Return the current process's effective group id.
*/

os.getegid=_.getegid;

/*
getgid() -> gid

Return the current process's group id.
*/

os.getgid=_.getgid;

/*
setgid(gid) -> None

Set the current process's group id.
*/

os.setgid=_.setgid;

/*
setpgid(pid, pgrp) -> None

Call the system call setpgid().
*/

os.setpgid=_.setpgid;

/*
seteuid(uid) -> None

Set the current process's effective user id.
*/

os.geteuid=_.geteuid;

/*
setpgrp() -> None

Make this process a session leader.
*/

os.setpgrp=_.setpgrp;

/*
setsid() -> None

Call the system call setsid().
*/

os.setsid=_.setsid;

/*
setuid(uid) -> None

Set the current process's user id.
*/

os.setuid=_.setuid;

/*
nice(inc) -> new_priority

Decrease the priority of process and return new priority.
*/

os.nice=_.nice;

/*
waitpid(pid, options) -> (pid, status)

Wait for completion of a give child process.
*/

os.waitpid=function(pid, options) {
	if (pid==undefined) pid=-1;
	if (options==undefined) options=0;
	var status=[0];
	var res=_.waitpid(pid,status,options);
	if (res==-1) os.error("os.waitpid");
	if (status[0] & 0x7f) {
		return [res,-(status[0] & 0x7f)]; // signaled;
	} else {
		return [res,(status[0] & 0xff00) >> 8];
	}
}

/*
wait() -> (pid, status)

Wait for completion of a child process.
*/

os.wait=os.waitpid;

/**/

os.spawn=function(type,args) {
	var pid=os.fork();
	if (pid) {
		// Parent process
		if (args[0]) { // wait
			return os.waitpid(pid);
		} else {
			return pid;
		}
	} else {
		func=os["exec"+type];
		var subargs=[];
		for (var i=1; i<args.length; i++) {
			subargs.push(args[i]);
		}
		func.apply(this,subargs);
	}
}

/*
spawnl(Boolean wait, String file, [String args, ...]) -> integer
 
Execute file with arguments from args in a subprocess.
If wait == true return the pid of the process.
If wait == false return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/


os.spawnl=function(wait, file) {
	return os.spawn.call(this,"l",arguments);
}

/*
spawnle(Boolean mode, String file, [String args, ...], Object env) -> integer
 
Execute file with arguments from args in a subprocess with the
supplied environment.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnle=function(wait, file) {
	return os.spawn.call(this,"le",arguments);
}

/*
spawnlp(Boolean mode, String file, [String args, ...], Object env) -> integer
 
Execute file (which is looked for along $PATH) with arguments from
args in a subprocess with the supplied environment.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnlp=function(wait, file) {
	return os.spawn.call(this,"lp",arguments);
}

/*
spawnlpe(Boolean mode, String file, [String args, ...] , Object env) -> integer
 
Execute file (which is looked for along $PATH) with arguments from
args in a subprocess with the supplied environment.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnlpe=function(wait, file) {
	return os.spawn.call(this,"lpe",arguments);
}

/*
spawnv(Boolean mode, String file, Array args) -> integer
 
Execute file with arguments from args in a subprocess.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnv=function(wait, file) {
	return os.spawn.call(this,"v",arguments);
}

/*
spawnvp(Boolean mode, String file, Array args) -> integer
 
Execute file (which is looked for along $PATH) with arguments from
args in a subprocess.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnvp=function(wait, file) {
	return os.spawn.call(this,"vp",arguments);
}

/*
spawnve(mode, file, args, env) -> integer
 
Execute file with arguments from args in a subprocess with the
specified environment.
If mode == P_NOWAIT return the pid of the process.
If mode == P_WAIT return the process's exit code if it exits normally;
otherwise return -SIG, where SIG is the signal that killed it.
*/

os.spawnve=function(wait, file) {
	return os.spawn.call(this,"ve",arguments);
}

/*
confstr(num) -> string

Return a string-valued system configuration variable.
*/


os.confstr=function(num) {
	var trylen=256;
	for (;;) {
		trylen*=2;
		var buf=new Pointer(trylen);
		var len=_.confstr(num, buf, trylen);
		if (len<trylen) break;
	}
	return buf.string(len);
}

/*
execl(String file, [String args...])
 
Execute the executable file with argument list args, replacing the
current process.
*/

os.execl=function(file) {
	var i;
	var args=[file,file];
	for (i=1; i<arguments.length; i++)
		args.push(String(arguments[i]));
	args.push(null);
	_.execl.apply(this,args);	
	os.error("os.execl: error executing '"+file+"'");
}

/*
execle(String file, [String args...], Object env)
 
Execute the executable file with argument list args
and environment env (object with property/value pairs), replacing the
current process.
*/

os.execle=function(file) {
	var i;
	var env=arguments[arguments.length-1];
	var envarg=[];
	var args=[file,file];
	for (i in env) {
		envarg.push(i+"="+String(env[i]));
	}
	envarg.push(null);
	for (i=1; i<arguments.length-1; i++)
		args.push(String(arguments[i]));
	args.push(null);
	args.push(envarg);
	_.execle.apply(this,args);	
	os.error("os.execle: error executing '"+file+"'");
}

/*
execlp(String file, [String args...])
 
Execute the executable file (which is searched for along PATH)
with argument list args, replacing the current process.
*/

os.execlp=function(file) {
	var i;
	var args=[file,file];
	for (i=1; i<arguments.length; i++)
		args.push(String(arguments[i]));
	args.push(null);
	_.execlp.apply(this,args);	
	os.error("os.execlp: error executing '"+file+"'");
}

/*
execlpe(String file, [String args...], Object env)

Execute the executable file (which is searched for along PATH) with argument list args
and environment env (object with property/value pairs), replacing the
current process.
*/

os.execlpe=function(file) {
	arguments[0]=os.which(arguments[0]);
	return os.execle.apply(this,arguments);
}


/*
execv(file, args)
 
Execute the executable file
with argument list args (an array), replacing the current process.
*/

os.execv=function(file,_args) {
	var i;
	var args=[file];
	for (i=0; i<_args.length; i++)
		args.push(String(_args[i]));
	args.push(null);
	_.execv.call(this,file,args);	
	os.error("os.execv: error executing '"+file+"'");
}

/*
execvp(String file, Array args)
 
Execute the executable file (which is searched for along PATH)
with argument list args (an array), replacing the current process.
*/

os.execvp=function(file,_args) {
	var i;
	var args=[file];
	for (i=0; i<_args.length; i++)
		args.push(String(_args[i]));
	args.push(null);
	_.execvp.call(this,file,args);	
	os.error("os.execvp: error executing '"+file+"'");
}


/*
execve(String file, Array args, Object env)
 
Execute the executable file
with argument list args (an array) and environment env (object with property/value pairs), replacing the current process.
*/

os.execve=function(file, _args, env) {
	var i;
	var env=arguments[arguments.length-1];
	var envarg=[];
	var args=[file,file];
	for (i in env) {
		envarg.push(i+"="+String(env[i]));
	}
	envarg.push(null);
	for (i=0; i<_args.length-1; i++)
		args.push(String(_args[i]));
	args.push(null);
	args.push(envarg);
	_.execle.apply(this,args);	
	os.error("os.execve: error executing '"+file+"'");
}


/*
ctermid() -> String

Return the name of the controlling terminal for this process.
*/

os.ctermid=function() {
	var buf=new Pointer(_.L_ctermid+1);
	_.ctermid(buf);
	return buf.string();
}

/*
====exists(path)====

Returns true if path exists.
Otherwise, returns false
*/

os.exists=os.access;


/*
getcwd() -> path

Return a string representing the current working directory.
*/

os.getcwd=function() {
	var trysize=256;
	for(;;) {
		var buf=new Pointer(trysize);
		if (_.getcwd(buf, trysize)!=null) break;
		trysize*=2;
	}
	return buf.string();
}

/*
os.getenv does not exist. Use the global ''environment'' variable instead.
*/

/*
os.getenv=function(key, dfault) {
	var env=_.getenv(key);
	if (env==null) return dfault;
	return env.string();
}
*/

/*
os.putenv does not exist. Use the global ''environment'' variable instead.
*/

/*
os.putenv=function(key, value) {
	if (_.putenv(key+"="+value)) os.error("os.putenv");
}
*/

/*
mkfifo(file, [mode=0666]) -> None

Create a FIFO (a POSIX named pipe).
*/

os.mkfifo=function(path, mode) {
	if (mode==undefined) mode=0666;
	return _.mkfifo(path,mode);
}

/*
mknod(filename, [mode=0600], device=[major,minor])
*/

os.mknod=function(filename, mode, device) {
	if (mode==undefined) mode=0600;
	return _.call_mknod(filename, mode, {__val:device});
}

/*
mkdir(path [, mode=0777]) -> None

Create a directory.
*/

os.mkdir=function(pathname, mode) {
	if (mode==undefined) mode=0777;
	_.mkdir(pathname, mode);
}

/*
makedirs(path [, mode=0777]) -> None
 
Super-mkdir; create a leaf directory and all intermediate ones.
Works like mkdir, except that any intermediate path segment (not
just the rightmost) will be created if it does not exist.  This is
recursive.
*/

os.makedirs=function(path, mode) {
	if (mode==undefined) mode=0777;

	var tomake=[];
	for(;;) {
		if (os.exists(path)) break;
		var lastslash=path.lastIndexOf(os.sep);
		if (lastslash==-1) os.error("os.makedirs");
		tomake.unshift(path.substr(lastslash+1));
		path=path.substr(0,lastslash);
	}
	while (tomake.length) {
		var thisdir=tomake.shift();
		_.mkdir(path+os.sep+thisdir, mode);
		path+=os.sep+thisdir;
	}
}

/*

*/

os.pathconf_names=function() {
	var ret=[];
	for (var i in C) {
		if (i.substr(0,4)=="_PC_") ret.push(i.substr(4));
	}
	return ret;
}

os.system=function(cmd) {
    var ret;
	if (ret=_.system(cmd)) throw(cmd+" returned "+ret);
}

/*
tmpfile() -> file object

Create a temporary file with no directory entries.
*/

os.tmpfile=function() {
	var fp=_.tmpfile();
	return new file(fp);
}

/*
tempnam([dir],[prefix]) -> string

Return a unique name for a temporary file.
*/

os.tempnam=function(dir, prefix) {
	if (dir==undefined) dir=null;
	if (prefix==undefined) prefix=null;

	var ret=_.tempnam(dir,prefix);
	if (ret==null) os.error("tempnam");
	var str=ret.string();
	_.free(ret);
	return str;
}

/*
tmpnam() -> string

Return a unique name for a temporary file.
*/

os.tmpnam=function() {
	var buf=new Pointer(_.L_tmpnam+1);
	_.tmpnam(buf);
	return buf.string();
}

/*
unlink(path) -> None

Remove a file (same as remove(path)).
*/

os.unlink=function(path) {
	if (_.unlink(path)==-1) throw("os: unlink '"+path+"'");
}

/*
remove(path) -> None

Remove a file (same as unlink(path)).
*/

os.remove=os.unlink;

/*
deltree(path) -> Array

Deletes a tree, returning an array with those files and directories
that could not be removed
*/

os.deltree=function(path, errors) {
	if (path[path.length-1]==os.sep) path=path.substr(0,path.length-1);
	if (errors==undefined) errors=[];
	if (os.isdir(path)) {
		for each (var filename in os.listdir(path)) {
			os.deltree(path+os.sep+filename,errors);
		}
		if (_.rmdir(path)==-1) errors.push(path);
	} else {
		if (_.unlink(path)==-1) errors.push(path);
	}
	return errors;
}

/*
rename(old, new) -> None

Rename a file or directory.
*/

os.rename=function(oldpath, newpath) {
	if (_.rename(oldpath, newpath)!=0) os.error("os.rename");
}

/*
renames(old, new) -> None
 
Super-rename; create directories as necessary and delete any left
empty.  Works like rename, except creation of any intermediate
directories needed to make the new pathname good is attempted
first.  After the rename, directories corresponding to rightmost
path segments of the old name will be pruned way until either the
whole path is consumed or a nonempty directory is found.
 
Note: this function can fail with the new directory structure made
if you lack permissions needed to unlink the leaf directory or
file.
*/

os.renames=function(old,nw) {
	var lastslash=nw.lastIndexOf(os.sep);
	if (lastslash!=-1) {
		os.makedirs(nw.substr(0,lastslash));
	}
	os.rename(old,nw);
}

/*
rmdir(path) -> None

Remove a directory.
*/

os.rmdir=function(path) {
	if (_.rmdir(path)==-1) os.error("os.rmdir");
}

/*
removedirs(path) -> None
 
Super-rmdir; remove a leaf directory and empty all intermediate
ones.  Works like rmdir except that, if the leaf directory is
successfully removed, directories corresponding to rightmost path
segments will be pruned way until either the whole path is
consumed or an error occurs.  Errors during this latter phase are
ignored -- they generally mean that a directory was not empty.

*/

os.removedirs=function(name) {
	for(;;) {
		try {
			os.rmdir(name);
		} catch (err) {
			break;
		}
		var slashpos=name.lastIndexOf(os.sep);
		if (slashpos==-1) break;
		name=name.substr(0,slashpos);
	}
}

/*
readlink(path) -> path

Return a string representing the path to which the symbolic link points.
*/

os.readlink=function(path) {
	var trysize=256;
	for(;;) {
		var buf=new Pointer(trysize);
		var len=readlink(path, buf, trysize);
		if (len<trysize) break;
		trysize*=2;
	}
	return buf.string(len);
}

/*
uname() -> (sysname, nodename, release, version, machine)

Return an object identifying the current operating system.
*/

os.uname=function() {
	var buf=_.New(_.struct.utsname);
	if (_.uname(buf)) {
		os.error("uname");
	}
	var i;
	var ret=buf.get();
	for (i in ret) {
		var str=String.fromCharCode.apply(String,ret[i]);
		var len=str.indexOf(String.fromCharCode(0));
		ret[i]=str.substr(0,len);
	}
	return ret;
}

/*
utime(path, [atime, [utime]]) -> None

Set the access and modified time of the file to the given values.  If no times
are given, set the access and modified times to the current time. Times should
be Date objects.
*/

os.utime=function(path, atime, utime) {
	var now=new Date;
	if (atime==undefined) atime=now;
	if (utime==undefined) utime=now;

	var tv_atime=_.New(_.struct.timeval);
	var tv_utime=_.New(_.struct.timeval);

	tv_atime.tv_sec=atime.getTime()/1000;
	tv_utime.tv_sec=utime.getTime()/1000;

	if (_.utimes(path, [tv_atime, tv_utime])) {
		os.error("utime");
	}
}

/*
which(name)

Searches for ''name'' in search path
*/

os.which=function(name) {
	if (name.indexOf(os.sep)!=-1) { // relative or absolute path, don't search
		if (os.access(name,os.X_OK)) return name;
		return false;
	}
	var path=environment.PATH.split(os.pathsep);
	for each(var p in path) {
		var tryname=p+os.sep+name;
		if (os.access(tryname,os.X_OK)) return tryname;
	}
	return false;
}

/*
touch(path)

Updates atime and mtime of path. Creates path if doesn't exist.

*/

os.touch=function(path) {
	var fd=_.open(path, _.O_CREAT | _.O_RDWR, 0777);
	if (fd==0)
		os.error("touch");
	_.close(fd);
}


os.curdir='.';
os.pardir='..';
